#!/usr/bin/env python
"""
Deploy all scripts and extensions for python2 installations.
Useless in python3.

This script is not meant to be called by users (it's automatically called by dirac-install).

Usage:
  dirac-deploy-scripts [options] ... ... <python path>

Arguments:
  python path:    you can specify the folder where your python installation should be fetched from
                  to replace the shebang

Example:
  $ dirac-deploy-scripts
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os
import re
import sys
import stat
import getopt
import shutil

cmdOpts = (
    ("", "symlink", "this will create symlinks instead of wrappers"),
    ("", "module=", "module in which to look for the scripts"),
    ("h", "help", "help doc string"),
)


def usage(err=""):
    """Usage printout

    :param str err: will print something like "option -a not recognized"
    """
    if err:
        print(err)
    print(__doc__)
    print("Options::\n\n")
    for cmdOpt in cmdOpts:
        print("  %s %s : %s" % (cmdOpt[0].ljust(3), cmdOpt[1].ljust(20), cmdOpt[2]))

    sys.exit(2 if err else 0)


DEBUG = False

moduleSuffix = "DIRAC"
gDefaultPerms = stat.S_IWUSR | stat.S_IRUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH
excludeMask = ["__init__.py"]
simpleCopyMask = [
    os.path.basename(__file__),
    "dirac-compile-externals.py",
    "dirac-install.py",
    "dirac-install-extension.py",
    "dirac-platform.py",
    "dirac_compile_externals.py",
    "dirac_install.py",
    "dirac_platform.py",
]

wrapperTemplate = """#!/bin/bash
source "$DIRAC/bashrc"
export DCOMMANDS_PPID=$PPID
exec $PYTHONLOCATION$ $DIRAC/$SCRIPTLOCATION$ "$@"
"""

# Python interpreter location can be specified as an argument
pythonLocation = "/usr/bin/env python"
# if True, do not use the script wrapper but just use symlinks
useSymlinks = False
# Module in which to look for the scripts
module = None

try:
    opts, args = getopt.getopt(sys.argv[1:], "".join([opt[0] for opt in cmdOpts]), [opt[1] for opt in cmdOpts])
except getopt.GetoptError as err:
    # print help information and exit
    usage(str(err))
for o, a in opts:
    if o in ("-h", "--help"):
        usage()
    if o == "--symlink":
        useSymlinks = True
    elif o == "--module":
        module = a
    else:
        assert False, "unhandled options %s" % o

if args:
    pythonLocation = os.path.join(args[0], "bin", "python")

wrapperTemplate = wrapperTemplate.replace("$PYTHONLOCATION$", pythonLocation)


def lookForScriptsInPath(basePath, rootModule):
    isScriptsDir = os.path.split(rootModule)[1] == "scripts"
    scriptFiles = []
    for entry in os.listdir(basePath):
        absEntry = os.path.join(basePath, entry)
        if os.path.isdir(absEntry):
            scriptFiles.extend(lookForScriptsInPath(absEntry, os.path.join(rootModule, entry)))
        elif isScriptsDir and os.path.isfile(absEntry):
            scriptFiles.append((os.path.join(rootModule, entry), entry))
    return scriptFiles


def findDIRACRoot(path):
    dirContents = os.listdir(path)
    if "DIRAC" in dirContents and os.path.isdir(os.path.join(path, "DIRAC")):
        return path
    parentPath = os.path.dirname(path)
    if parentPath == path or len(parentPath) == 1:
        return False
    return findDIRACRoot(os.path.dirname(path))


rootPath = findDIRACRoot(os.path.dirname(os.path.realpath(__file__)))
if not rootPath:
    print("Error: Cannot find DIRAC root!")
    sys.exit(1)

# hack alert!
# Removing 'DIRAC/src/', which is added when installing py2 releases with dirac-install and "-m" flag
rootPath = rootPath.replace("DIRAC/src/", "")

targetScriptsPath = os.path.join(rootPath, "scripts")
pythonScriptRE = re.compile("(.*/)*([a-z]+-[a-zA-Z0-9-]+|[a-z]+_[a-zA-Z0-9_]+|d[a-zA-Z0-9-]+).py$")
print("Scripts will be deployed at %s" % targetScriptsPath)

if not os.path.isdir(targetScriptsPath):
    os.mkdir(targetScriptsPath)

if module:
    listDir = module.split(",")
else:
    listDir = os.listdir(rootPath)
    # DIRAC scripts need to be treated first, so that its scripts
    # can be overwritten by the extensions
    if "DIRAC" in listDir:  # should always be true...
        listDir.remove("DIRAC")
        listDir.insert(0, "DIRAC")

for rootModule in listDir:
    modulePath = os.path.join(rootPath, rootModule)
    if not os.path.isdir(modulePath):
        continue
    extSuffixPos = rootModule.find(moduleSuffix)
    if extSuffixPos == -1 or extSuffixPos != len(rootModule) - len(moduleSuffix):
        continue
    print(("Inspecting %s module" % rootModule))
    scripts = lookForScriptsInPath(modulePath, rootModule)
    for script in scripts:
        scriptPath = script[0]
        scriptName = script[1]
        if scriptName in excludeMask:
            continue
        scriptLen = len(scriptName)
        if scriptName not in simpleCopyMask and pythonScriptRE.match(scriptName):
            newScriptName = scriptName[:-3].replace("_", "-")
            if DEBUG:
                print((" Wrapping %s as %s" % (scriptName, newScriptName)))
            fakeScriptPath = os.path.join(targetScriptsPath, newScriptName)

            # Either create the symlink or write the wrapper script
            if useSymlinks:
                # We may overwrite already existing links (in extension for example)
                # os.symlink will not allow that, so remove the existing first
                if os.path.exists(fakeScriptPath):
                    os.remove(fakeScriptPath)
                # Create the symlink
                os.symlink(os.path.join(rootPath, scriptPath), fakeScriptPath)
            else:
                with open(fakeScriptPath, "w") as fd:
                    fd.write(wrapperTemplate.replace("$SCRIPTLOCATION$", scriptPath))

            os.chmod(fakeScriptPath, gDefaultPerms)
        else:
            if DEBUG:
                print((" Copying %s" % scriptName))
            shutil.copy(os.path.join(rootPath, scriptPath), targetScriptsPath)
            copyPath = os.path.join(targetScriptsPath, scriptName)
            os.chmod(copyPath, gDefaultPerms)
            cLen = len(copyPath)
            reFound = pythonScriptRE.match(copyPath)
            if reFound:
                pathList = list(reFound.groups())
                pathList[-1] = pathList[-1].replace("_", "-")
                destPath = "".join(pathList)
                if DEBUG:
                    print((" Renaming %s as %s" % (copyPath, destPath)))
                os.rename(copyPath, destPath)



""" EmailAgent
  This agent reads the ResourceStatusCache table of ResourceStatusDB
  for sending emails with aggregated information about state changes,
  and then clears it.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN EmailAgent
  :end-before: ##END
  :dedent: 2
  :caption: EmailAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gConfig, S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ResourceStatusSystem.Utilities import RssConfiguration
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

__RCSID__ = "$Id$"

AGENT_NAME = "ResourceStatus/EmailAgent"


class EmailAgent(AgentModule):
    def __init__(self, *args, **kwargs):

        AgentModule.__init__(self, *args, **kwargs)
        self.diracAdmin = None
        self.default_value = None

        self.rsClient = ResourceStatusClient()

    def initialize(self, *args, **kwargs):
        """EmailAgent initialization"""

        self.diracAdmin = DiracAdmin()

        return S_OK()

    @staticmethod
    def _groupBySiteName(result):
        """
        Group results by SiteName
        """

        siteNameCol = result["Columns"].index("SiteName")
        resultValue = result["Value"]

        siteNameDict = {}
        for row in resultValue:
            if row[siteNameCol] not in siteNameDict:
                siteNameDict[row[siteNameCol]] = [row]
            else:
                siteNameDict[row[siteNameCol]].append(row)

        return siteNameDict

    def execute(self):

        result = self.rsClient.select("ResourceStatusCache")
        if not result["OK"]:
            return S_ERROR()

        columnNames = result["Columns"]
        result = self._groupBySiteName(result)

        for site, records in result.items():

            email = ""
            html_body = ""
            html_elements = ""

            if gConfig.getValue("/DIRAC/Setup"):
                setup = "(" + gConfig.getValue("/DIRAC/Setup") + ")\n\n"
            else:
                setup = ""

            html_header = """\
      <!DOCTYPE html>
      <html>
      <head>
      <meta charset='UTF-8'>
        <style>
          table{{color:#333;font-family:Helvetica,Arial,sans-serif;min-width:700px;border-collapse:collapse;border-spacing:0}}
          td,th{{border:1px solid transparent;height:30px;transition:all .3s}}th{{background:#DFDFDF;font-weight:700}}
          td{{background:#FAFAFA;text-align:center}}.setup{{font-size:150%;color:grey}}.Banned{{color:red}}.Error{{color:#8b0000}}
          .Degraded{{color:gray}}.Probing{{color:#00f}}.Active{{color:green}}tr:nth-child(even)
          td{{background:#F1F1F1}}tr:nth-child(odd)
          td{{background:#FEFEFE}}tr td:hover{{background:#666;color:#FFF}}
        </style>
      </head>
      <body>
        <p class="setup">{setup}</p>
      """.format(
                setup=setup
            )

            for row in records:
                statusType = row[columnNames.index("StatusType")]
                resourceName = row[columnNames.index("ResourceName")]
                status = row[columnNames.index("Status")]
                time = row[columnNames.index("Time")]
                previousStatus = row[columnNames.index("PreviousStatus")]
                html_elements += (
                    "<tr>"
                    + "<td>"
                    + statusType
                    + "</td>"
                    + "<td>"
                    + resourceName
                    + "</td>"
                    + "<td class='"
                    + status
                    + "'>"
                    + status
                    + "</td>"
                    + "<td>"
                    + str(time)
                    + "</td>"
                    + "<td class='"
                    + previousStatus
                    + "'>"
                    + previousStatus
                    + "</td>"
                    + "</tr>"
                )

            html_body = """\
        <table>
          <tr>
              <th>Status Type</th>
              <th>Resource Name</th>
              <th>Status</th>
              <th>Time</th>
              <th>Previous Status</th>
          </tr>
          {html_elements}
        </table>
      </body>
      </html>
      """.format(
                html_elements=html_elements
            )

            email = html_header + html_body

            subject = "RSS actions taken for " + str(site) + "\n"
            self._sendMail(subject, email, html=True)

        self.rsClient.delete("ResourceStatusCache")

        return S_OK()

    def _sendMail(self, subject, body, html=False):

        userEmails = self._getUserEmails()
        if not userEmails["OK"]:
            return userEmails

        # User email address used to send the emails from.
        fromAddress = RssConfiguration.RssConfiguration().getConfigFromAddress()

        for user in userEmails["Value"]:

            # FIXME: should not I get the info from the RSS User cache ?

            resEmail = self.diracAdmin.sendMail(user, subject, body, fromAddress=fromAddress, html=html)
            if not resEmail["OK"]:
                return S_ERROR('Cannot send email to user "%s"' % user)

        return S_OK()

    def _getUserEmails(self):

        configResult = RssConfiguration.getnotificationGroups()
        if not configResult["OK"]:
            return configResult
        try:
            notificationGroups = configResult["Value"]["notificationGroups"]
        except KeyError:
            return S_ERROR("%s/notificationGroups not found")

        notifications = RssConfiguration.getNotifications()
        if not notifications["OK"]:
            return notifications
        notifications = notifications["Value"]

        userEmails = []

        for notificationGroupName in notificationGroups:
            try:
                userEmails.extend(notifications[notificationGroupName]["users"])
            except KeyError:
                self.log.error("%s not present" % notificationGroupName)

        return S_OK(userEmails)



""" Test class for plugins
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

# imports
import unittest

# sut
from DIRAC.Core.Utilities.Time import timeThis


class logClass(object):
    def __init__(self):
        self._systemName = "aSystemName"
        self._subName = "sSubName"


@timeThis
def myMethod():
    print("boh")


class myClass(object):
    def __init__(self):
        self.log = logClass()

    @timeThis
    def myMethodInAClass(self):
        print("boh")


class myBetterClass(object):
    def __init__(self):
        self.log = logClass()
        self.log._subName = "anotherSubName"

    @timeThis
    def myMethodInAClass(self):
        print("boh")


class myEvenBetterClass(object):
    def __init__(self):
        self.log = logClass()
        self.transString = "this is a transString"

    @timeThis
    def myMethodInAClass(self, a, b=None):
        print("boh")


class TimeTestCase(unittest.TestCase):
    """Base class for the Agents test cases"""

    def setUp(self):
        pass

    def tearDown(self):
        pass


class TimeSuccess(TimeTestCase):
    def test_timeThis(self):

        self.assertIsNone(myMethod())
        self.assertIsNone(myClass().myMethodInAClass())
        self.assertIsNone(myBetterClass().myMethodInAClass())
        a1 = ["aa", "bb"]
        a2 = "bb"
        self.assertIsNone(myEvenBetterClass().myMethodInAClass(a1, b=a2))
        a1 = "aa"
        a2 = {"a": "aa", "b": "bb"}
        self.assertIsNone(myEvenBetterClass().myMethodInAClass(a1, b=a2))
        a1 = "aa"
        a2 = {"a": "aa", "b": "bb"}
        self.assertIsNone(myEvenBetterClass().myMethodInAClass(a=a1, b=a2))
        a1 = "aa"
        a2 = {"a": "aa", "b": "bb", "c": "cc"}
        self.assertIsNone(myEvenBetterClass().myMethodInAClass(a=a2, b=a2))


#############################################################################
# Test Suite run
#############################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TimeTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TimeSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import re
import time
import gzip
from six.moves import queue as Queue
import threading
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Core.Utilities.File import mkDir


class SecurityFileLog(threading.Thread):
    def __init__(self, basePath, daysToLog=100):
        self.__basePath = basePath
        self.__messagesQueue = Queue.Queue()
        self.__requiredFields = (
            "timestamp",
            "success",
            "sourceIP",
            "sourcePort",
            "sourceIdentity",
            "destinationIP",
            "destinationPort",
            "destinationService",
            "action",
        )
        threading.Thread.__init__(self)
        self.__secsToLog = daysToLog * 86400
        gThreadScheduler.addPeriodicTask(
            86400, self.__launchCleaningOldLogFiles, elapsedTime=(time.time() % 86400) + 3600
        )
        self.setDaemon(True)
        self.start()

    def run(self):
        while True:
            secMsg = self.__messagesQueue.get()
            msgTime = secMsg[0]
            path = "%s/%s/%02d" % (self.__basePath, msgTime.year, msgTime.month)
            mkDir(path)
            logFile = "%s/%s%02d%02d.security.log.csv" % (path, msgTime.year, msgTime.month, msgTime.day)
            if not os.path.isfile(logFile):
                fd = open(logFile, "w")
                fd.write(
                    "Time, Success, Source IP, Source Port, source Identity, destinationIP,\
           destinationPort, destinationService, action\n"
                )
            else:
                fd = open(logFile, "a")
            fd.write("%s\n" % ", ".join([str(item) for item in secMsg]))
            fd.close()

    def __launchCleaningOldLogFiles(self):
        nowEpoch = time.time()
        self.__walkOldLogs(
            self.__basePath, nowEpoch, re.compile(r"^\d*\.security\.log\.csv$"), 86400 * 3, self.__zipOldLog
        )
        self.__walkOldLogs(
            self.__basePath,
            nowEpoch,
            re.compile(r"^\d*\.security\.log\.csv\.gz$"),
            self.__secsToLog,
            self.__unlinkOldLog,
        )

    def __unlinkOldLog(self, filePath):
        try:
            gLogger.info("Unlinking file %s" % filePath)
            os.unlink(filePath)
        except Exception as e:
            gLogger.error("Can't unlink old log file", "%s: %s" % (filePath, str(e)))
            return 1
        return 0

    def __zipOldLog(self, filePath):
        try:
            gLogger.info("Compressing file %s" % filePath)
            fd = gzip.open("%s.gz" % filePath, "w")
            with open(filePath) as fO:
                bS = 1048576  # 1MiB
                buf = fO.read(bS)
                while buf:
                    fd.write(buf)
                    buf = fO.read(bS)
            fd.close()
        except Exception as e:
            gLogger.exception("Can't compress old log file", filePath)
            return 1
        return self.__unlinkOldLog(filePath) + 1

    def __walkOldLogs(self, path, nowEpoch, reLog, executionInSecs, functor):
        initialEntries = os.listdir(path)
        files = []
        numEntries = 0
        for entry in initialEntries:
            entryPath = os.path.join(path, entry)
            if os.path.isdir(entryPath):
                numEntries += 1
                numEntriesSubDir = self.__walkOldLogs(entryPath, nowEpoch, reLog, executionInSecs, functor)
                if numEntriesSubDir == 0:
                    gLogger.info("Removing dir %s" % entryPath)
                    try:
                        os.rmdir(entryPath)
                        numEntries -= 1
                    except Exception as e:
                        gLogger.error("Can't delete directory", "%s: %s" % (entryPath, str(e)))
            elif os.path.isfile(entryPath):
                numEntries += 1
                if reLog.match(entry):
                    if nowEpoch - os.stat(entryPath)[8] > executionInSecs:
                        numEntries += functor(entryPath) - 1
        return numEntries

    def logAction(self, msg):
        if len(msg) != len(self.__requiredFields):
            return S_ERROR(
                "Mismatch in the msg size, it should be %s and it's %s" % (len(self.__requiredFields), len(msg))
            )
        self.__messagesQueue.put(msg)
        return S_OK()
